@@ -1,8 +1,11 @@
 from __future__ import with_statement
+from itertools import starmap
 import datetime
 import time
 import warnings
-from itertools import imap, izip, starmap
+
+from redis._compat import (b, izip, imap, iteritems, dictkeys, dictvalues,
+                           basestring, long, nativestr)
 from redis.connection import ConnectionPool, UnixDomainSocketConnection
 from redis.exceptions import (
     ConnectionError,
@@ -12,6 +15,9 @@ from redis.exceptions import (
     WatchError,
 )
 
+SYM_EMPTY = b('')
+
+
 def list_or_args(keys, args):
     # returns a single list combining keys and args
     try:
@@ -26,6 +32,7 @@ def list_or_args(keys, args):
         keys.extend(args)
     return keys
 
+
 def timestamp_to_datetime(response):
     "Converts a unix timestamp to a Python datetime object"
     if not response:
@@ -36,14 +43,17 @@ def timestamp_to_datetime(response):
         return None
     return datetime.datetime.fromtimestamp(response)
 
+
 def string_keys_to_dict(key_string, callback):
     return dict.fromkeys(key_string.split(), callback)
 
+
 def dict_merge(*dicts):
     merged = {}
     [merged.update(d) for d in dicts]
     return merged
 
+
 def parse_debug_object(response):
     "Parse the results of Redis's DEBUG OBJECT command into a Python dict"
     # The 'type' of the object is the first item in the response, but isn't
@@ -60,12 +70,14 @@ def parse_debug_object(response):
 
     return response
 
+
 def parse_object(response, infotype):
     "Parse the results of an OBJECT command"
     if infotype in ('idletime', 'refcount'):
         return int(response)
     return response
 
+
 def parse_info(response):
     "Parse the result of Redis's INFO command into a Python dict"
     info = {}
@@ -95,11 +107,13 @@ def parse_info(response):
                 info[key] = get_value(value)
     return info
 
+
 def pairs_to_dict(response):
     "Create a dict given a list of key/value pairs"
     it = iter(response)
     return dict(izip(it, it))
 
+
 def zset_score_pairs(response, **options):
     """
     If ``withscores`` is specified in the options, return the response as
@@ -109,24 +123,28 @@ def zset_score_pairs(response, **options):
         return response
     score_cast_func = options.get('score_cast_func', float)
     it = iter(response)
-    return zip(it, imap(score_cast_func, it))
+    return list(izip(it, imap(score_cast_func, it)))
+
 
 def int_or_none(response):
     if response is None:
         return None
     return int(response)
 
+
 def float_or_none(response):
     if response is None:
         return None
     return float(response)
 
+
 def parse_config(response, **options):
     # this is stupid, but don't have a better option right now
     if options['parse'] == 'GET':
         return response and pairs_to_dict(response) or {}
     return response == 'OK'
 
+
 class StrictRedis(object):
     """
     Implementation of the Redis protocol.
@@ -142,35 +160,38 @@ class StrictRedis(object):
             'AUTH DEL EXISTS EXPIRE EXPIREAT HDEL HEXISTS HMSET MOVE MSETNX '
             'PERSIST RENAMENX SISMEMBER SMOVE SETEX SETNX SREM ZREM',
             bool
-            ),
+        ),
         string_keys_to_dict(
             'DECRBY GETBIT HLEN INCRBY LINSERT LLEN LPUSHX RPUSHX SADD SCARD '
             'SDIFFSTORE SETBIT SETRANGE SINTERSTORE STRLEN SUNIONSTORE ZADD '
             'ZCARD ZREMRANGEBYRANK ZREMRANGEBYSCORE',
             int
-            ),
+        ),
         string_keys_to_dict(
             # these return OK, or int if redis-server is >=1.3.4
             'LPUSH RPUSH',
-            lambda r: isinstance(r, long) and r or r == 'OK'
-            ),
+            lambda r: isinstance(r, long) and r or nativestr(r) == 'OK'
+        ),
         string_keys_to_dict('ZSCORE ZINCRBY', float_or_none),
         string_keys_to_dict(
             'FLUSHALL FLUSHDB LSET LTRIM MSET RENAME '
             'SAVE SELECT SET SHUTDOWN SLAVEOF WATCH UNWATCH',
-            lambda r: r == 'OK'
-            ),
+            lambda r: nativestr(r) == 'OK'
+        ),
         string_keys_to_dict('BLPOP BRPOP', lambda r: r and tuple(r) or None),
-        string_keys_to_dict('SDIFF SINTER SMEMBERS SUNION',
+        string_keys_to_dict(
+            'SDIFF SINTER SMEMBERS SUNION',
             lambda r: r and set(r) or set()
-            ),
-        string_keys_to_dict('ZRANGE ZRANGEBYSCORE ZREVRANGE ZREVRANGEBYSCORE',
+        ),
+        string_keys_to_dict(
+            'ZRANGE ZRANGEBYSCORE ZREVRANGE ZREVRANGEBYSCORE',
             zset_score_pairs
-            ),
+        ),
         string_keys_to_dict('ZRANK ZREVRANK', int_or_none),
         {
-            'BGREWRITEAOF': lambda r: \
-                r == 'Background rewriting of AOF file started',
+            'BGREWRITEAOF': (
+                lambda r: r == 'Background rewriting of AOF file started'
+            ),
             'BGSAVE': lambda r: r == 'Background saving started',
             'BRPOPLPUSH': lambda r: r and r or None,
             'CONFIG': parse_config,
@@ -179,10 +200,10 @@ class StrictRedis(object):
             'INFO': parse_info,
             'LASTSAVE': timestamp_to_datetime,
             'OBJECT': parse_object,
-            'PING': lambda r: r == 'PONG',
+            'PING': lambda r: nativestr(r) == 'PONG',
             'RANDOMKEY': lambda r: r and r or None,
         }
-        )
+    )
 
     def __init__(self, host='localhost', port=6379,
                  db=0, password=None, socket_timeout=None,
@@ -197,7 +218,7 @@ class StrictRedis(object):
                 'encoding': charset,
                 'encoding_errors': errors,
                 'decode_responses': decode_responses,
-                }
+            }
             # based on input, setup appropriate connection args
             if unix_socket_path:
                 kwargs.update({
@@ -478,7 +499,7 @@ class StrictRedis(object):
     def mset(self, mapping):
         "Sets each key in the ``mapping`` dict to its corresponding value"
         items = []
-        for pair in mapping.iteritems():
+        for pair in iteritems(mapping):
             items.extend(pair)
         return self.execute_command('MSET', *items)
 
@@ -488,7 +509,7 @@ class StrictRedis(object):
         none of the keys are already set
         """
         items = []
-        for pair in mapping.iteritems():
+        for pair in iteritems(mapping):
             items.extend(pair)
         return self.execute_command('MSETNX', *items)
 
@@ -583,7 +604,8 @@ class StrictRedis(object):
         """
         Unwatches the value at key ``name``, or None of the key doesn't exist
         """
-        warnings.warn(DeprecationWarning('Call UNWATCH from a Pipeline object'))
+        warnings.warn(
+            DeprecationWarning('Call UNWATCH from a Pipeline object'))
 
     #### LIST COMMANDS ####
     def blpop(self, keys, timeout=0):
@@ -783,7 +805,6 @@ class StrictRedis(object):
             pieces.append(store)
         return self.execute_command('SORT', *pieces)
 
-
     #### SET COMMANDS ####
     def sadd(self, name, *values):
         "Add ``value(s)`` to set ``name``"
@@ -856,7 +877,6 @@ class StrictRedis(object):
         args = list_or_args(keys, args)
         return self.execute_command('SUNIONSTORE', dest, *args)
 
-
     #### SORTED SET COMMANDS ####
     def zadd(self, name, *args, **kwargs):
         """
@@ -875,7 +895,7 @@ class StrictRedis(object):
                 raise RedisError("ZADD requires an equal number of "
                                  "values and scores")
             pieces.extend(args)
-        for pair in kwargs.iteritems():
+        for pair in iteritems(kwargs):
             pieces.append(pair[1])
             pieces.append(pair[0])
         return self.execute_command('ZADD', name, *pieces)
@@ -915,15 +935,17 @@ class StrictRedis(object):
         ``score_cast_func`` a callable used to cast the score return value
         """
         if desc:
-            return self.zrevrange(name, start, end, withscores, score_cast_func)
+            return self.zrevrange(name, start, end, withscores,
+                                  score_cast_func)
         pieces = ['ZRANGE', name, start, end]
         if withscores:
             pieces.append('withscores')
-        options = {'withscores': withscores, 'score_cast_func': score_cast_func}
+        options = {
+            'withscores': withscores, 'score_cast_func': score_cast_func}
         return self.execute_command(*pieces, **options)
 
-    def zrangebyscore(self, name, min, max,
-            start=None, num=None, withscores=False, score_cast_func=float):
+    def zrangebyscore(self, name, min, max, start=None, num=None,
+                      withscores=False, score_cast_func=float):
         """
         Return a range of values from the sorted set ``name`` with scores
         between ``min`` and ``max``.
@@ -944,7 +966,8 @@ class StrictRedis(object):
             pieces.extend(['LIMIT', start, num])
         if withscores:
             pieces.append('withscores')
-        options = {'withscores': withscores, 'score_cast_func': score_cast_func}
+        options = {
+            'withscores': withscores, 'score_cast_func': score_cast_func}
         return self.execute_command(*pieces, **options)
 
     def zrank(self, name, value):
@@ -990,11 +1013,12 @@ class StrictRedis(object):
         pieces = ['ZREVRANGE', name, start, num]
         if withscores:
             pieces.append('withscores')
-        options = {'withscores': withscores, 'score_cast_func': score_cast_func}
+        options = {
+            'withscores': withscores, 'score_cast_func': score_cast_func}
         return self.execute_command(*pieces, **options)
 
-    def zrevrangebyscore(self, name, max, min,
-            start=None, num=None, withscores=False, score_cast_func=float):
+    def zrevrangebyscore(self, name, max, min, start=None, num=None,
+                         withscores=False, score_cast_func=float):
         """
         Return a range of values from the sorted set ``name`` with scores
         between ``min`` and ``max`` in descending order.
@@ -1015,7 +1039,8 @@ class StrictRedis(object):
             pieces.extend(['LIMIT', start, num])
         if withscores:
             pieces.append('withscores')
-        options = {'withscores': withscores, 'score_cast_func': score_cast_func}
+        options = {
+            'withscores': withscores, 'score_cast_func': score_cast_func}
         return self.execute_command(*pieces, **options)
 
     def zrevrank(self, name, value):
@@ -1040,7 +1065,7 @@ class StrictRedis(object):
     def _zaggregate(self, command, dest, keys, aggregate=None):
         pieces = [command, dest, len(keys)]
         if isinstance(keys, dict):
-            keys, weights = keys.keys(), keys.values()
+            keys, weights = dictkeys(keys), dictvalues(keys)
         else:
             weights = None
         pieces.extend(keys)
@@ -1103,7 +1128,7 @@ class StrictRedis(object):
         if not mapping:
             raise DataError("'hmset' with 'mapping' of length 0")
         items = []
-        for pair in mapping.iteritems():
+        for pair in iteritems(mapping):
             items.extend(pair)
         return self.execute_command('HMSET', name, *items)
 
@@ -1201,7 +1226,7 @@ class Redis(StrictRedis):
                 raise RedisError("ZADD requires an equal number of "
                                  "values and scores")
             pieces.extend(reversed(args))
-        for pair in kwargs.iteritems():
+        for pair in iteritems(kwargs):
             pieces.append(pair[1])
             pieces.append(pair[0])
         return self.execute_command('ZADD', name, *pieces)
@@ -1224,7 +1249,7 @@ class PubSub(object):
         self.subscription_count = 0
         self.subscribe_commands = set(
             ('subscribe', 'psubscribe', 'unsubscribe', 'punsubscribe')
-            )
+        )
 
     def __del__(self):
         try:
@@ -1257,7 +1282,7 @@ class PubSub(object):
             self.connection = self.connection_pool.get_connection(
                 'pubsub',
                 self.shard_hint
-                )
+            )
         connection = self.connection
         try:
             connection.send_command(*args)
@@ -1477,15 +1502,16 @@ class BasePipeline(object):
         return self
 
     def _execute_transaction(self, connection, commands):
-        all_cmds = ''.join(starmap(connection.pack_command,
-                                   [args for args, options in commands]))
+        all_cmds = SYM_EMPTY.join(
+            starmap(connection.pack_command,
+                    [args for args, options in commands]))
         connection.send_packed_command(all_cmds)
         # we don't care about the multi/exec any longer
         commands = commands[1:-1]
         # parse off the response for MULTI and all commands prior to EXEC.
         # the only data we care about is the response the EXEC
         # which is the last command
-        for i in range(len(commands)+1):
+        for i in range(len(commands) + 1):
             self.parse_response(connection, '_')
         # parse the EXEC.
         response = self.parse_response(connection, '_')
@@ -1509,8 +1535,9 @@ class BasePipeline(object):
 
     def _execute_pipeline(self, connection, commands):
         # build up all commands into a single request to increase network perf
-        all_cmds = ''.join(starmap(connection.pack_command,
-                                   [args for args, options in commands]))
+        all_cmds = SYM_EMPTY.join(
+            starmap(connection.pack_command,
+                    [args for args, options in commands]))
         connection.send_packed_command(all_cmds)
         return [self.parse_response(connection, args[0], **options)
                 for args, options in commands]
@@ -1535,7 +1562,8 @@ class BasePipeline(object):
 
         conn = self.connection
         if not conn:
-            conn = self.connection_pool.get_connection('MULTI', self.shard_hint)
+            conn = self.connection_pool.get_connection('MULTI',
+                                                       self.shard_hint)
             # assign to self.connection so reset() releases the connection
             # back to the pool after we're done
             self.connection = conn
@@ -1544,11 +1572,11 @@ class BasePipeline(object):
             return execute(conn, stack)
         except ConnectionError:
             conn.disconnect()
-            # if we were watching a variable, the watch is no longer valid since
-            # this connection has died. raise a WatchError, which indicates
-            # the user should retry his transaction. If this is more than a
-            # temporary failure, the WATCH that the user next issue will fail,
-            # propegating the real ConnectionError
+            # if we were watching a variable, the watch is no longer valid
+            # since this connection has died. raise a WatchError, which
+            # indicates the user should retry his transaction. If this is more
+            # than a temporary failure, the WATCH that the user next issue
+            # will fail, propegating the real ConnectionError
             if self.watching:
                 raise WatchError("A ConnectionError occured on while watching "
                                  "one or more keys")
@@ -1577,14 +1605,17 @@ class StrictPipeline(BasePipeline, StrictRedis):
     "Pipeline for the StrictRedis class"
     pass
 
+
 class Pipeline(BasePipeline, Redis):
     "Pipeline for the Redis class"
     pass
 
+
 class LockError(RedisError):
     "Errors thrown from the Lock"
     pass
 
+
 class Lock(object):
     """
     A shared, distributed Lock. Using Redis for locking allows the Lock
@@ -1594,7 +1625,7 @@ class Lock(object):
     multiple clients play nicely together.
     """
 
-    LOCK_FOREVER = float(2**31+1) # 1 past max unix time
+    LOCK_FOREVER = float(2 ** 31 + 1)  # 1 past max unix time
 
     def __init__(self, redis, name, timeout=None, sleep=0.1):
         """
@@ -1609,8 +1640,8 @@ class Lock(object):
         holding the lock.
 
         Note: If using ``timeout``, you should make sure all the hosts
-        that are running clients have their time synchronized with a network time
-        service like ntp.
+        that are running clients have their time synchronized with a network
+        time service like ntp.
         """
         self.redis = redis
         self.name = name
